# coding=utf-8
# @license
# Copyright 2019-2020 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wrappers for representing a Neuroglancer coordinate space."""

import collections
import numpy as np

__all__ = []


def export(obj):
    __all__.append(obj.__name__)
    return obj


si_prefixes = {
    'Y': 24,
    'Z': 21,
    'E': 18,
    'P': 15,
    'T': 12,
    'G': 9,
    'M': 6,
    'k': 3,
    'h': 2,
    '': 0,
    'c': -2,
    'm': -3,
    'u': -6,
    'µ': -6,
    'n': -9,
    'p': -12,
    'f': -15,
    'a': -18,
    'z': -21,
    'y': -24,
}

si_units = ['m', 's', 'rad/s', 'Hz']

si_units_with_prefixes = {
    '%s%s' % (prefix, unit): (unit, exponent)
    for (prefix, exponent) in si_prefixes.items() for unit in si_units
}

si_units_with_prefixes[''] = ('', 0)


def parse_unit(scale, unit):
    unit, exponent = si_units_with_prefixes[unit]
    if exponent >= 0:
        return (scale * 10**exponent, unit)
    else:
        return (scale / 10**(-exponent), unit)


@export
class DimensionScale(collections.namedtuple('DimensionScale', ['scale', 'unit'])):
    __slots__ = ()

    def __new__(cls, scale=1, unit=''):
        return super(DimensionScale, cls).__new__(cls, scale, unit)


@export
class CoordinateSpace(object):
    __slots__ = ('names', 'scales', 'units')

    def __init__(self, json=None, names=None, scales=None, units=None):
        if json is None:
            if names is not None:
                self.names = tuple(names)
                scales = np.array(scales, dtype=np.float64)
                if isinstance(units, str):
                    units = tuple(units for _ in names)
                scales_and_units = tuple(
                    parse_unit(scale, unit) for scale, unit in zip(scales, units))
                scales = np.array([s[0] for s in scales_and_units], dtype=np.float64)
                units = tuple(s[1] for s in scales_and_units)
                self.units = units
                self.scales = scales
            else:
                self.names = ()
                self.scales = np.zeros(0, dtype=np.float64)
                self.units = ()
        else:
            if not isinstance(json, dict): raise TypeError
            self.names = tuple(json.keys())
            self.scales = np.array([json[k][0] for k in self.names], dtype=np.float64)
            self.units = tuple(json[k][1] for k in self.names)
        self.scales.setflags(write=False)

    @property
    def rank(self):
        return len(self.names)

    def __getitem__(self, i):
        if isinstance(i, str):
            idx = self.names.index(i)
            return DimensionScale(scale=self.scales[idx], unit=self.units[idx])
        if isinstance(i, slice):
            idxs = range(self.rank)[i]
            return [DimensionScale(scale=self.scales[j], unit=self.units[j]) for j in idxs]
        return DimensionScale(scale=self.scales[i], unit=self.units[i])

    def __repr__(self):
        return 'CoordinateSpace(%r)' % (self.to_json(), )

    def to_json(self):
        d = collections.OrderedDict()
        for name, scale, unit in zip(self.names, self.scales, self.units):
            d[name] = [scale, unit]
        return d



# coding=utf-8
# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wrappers for representing the Neuroglancer viewer state."""

from __future__ import absolute_import

import collections
import copy
import math
import numbers

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc

import numpy as np
import six

from . import local_volume
from . import skeleton
from . import segment_colors
from .coordinate_space import DimensionScale, CoordinateSpace
from .equivalence_map import EquivalenceMap
from .json_utils import encode_json_for_repr
from .json_wrappers import (JsonObjectWrapper, array_wrapper, optional, text_type, typed_list,
                            typed_map, typed_set, typed_string_map, wrapped_property,
                            number_or_string)

__all__ = ['CoordinateSpace', 'DimensionScale']


def export(obj):
    __all__.append(obj.__name__)
    return obj


def interpolate_linear(a, b, t):
    return a * (1 - t) + b * t

def interpolate_linear_optional_vectors(a, b, t):
    if a is not None and b is not None and len(a) == len(b):
        return a * (1 - t) + b * t
    return a

def unit_quaternion():
    return np.array([0, 0, 0, 1], np.float32)


def quaternion_slerp(a, b, t):
    """Spherical linear interpolation for unit quaternions.

    This is based on the implementation in the gl-matrix package:
    https://github.com/toji/gl-matrix
    """
    if a is None:
        a = unit_quaternion()
    if b is None:
        b = unit_quaternion()
    # calc cosine
    cosom = np.dot(a, b)
    # adjust signs (if necessary)
    if cosom < 0.0:
        cosom = -cosom
        b = -b

    # calculate coefficients
    if (1.0 - cosom) > 0.000001:
        # standard case (slerp)
        omega = math.acos(cosom)
        sinom = math.sin(omega)
        scale0 = math.sin((1.0 - t) * omega) / sinom
        scale1 = math.sin(t * omega) / sinom
    else:
        # "from" and "to" quaternions are very close
        #  ... so we can do a linear interpolation
        scale0 = 1.0 - t
        scale1 = t
    return scale0 * a + scale1 * b


def interpolate_zoom(a, b, t):
    if a is None or b is None:
        return a
    scale_change = math.log(b / a)
    return a * math.exp(scale_change * t)


@export
class Tool(JsonObjectWrapper):
    __slots__ = ()

    type = wrapped_property('type', text_type)

    def __init__(self, json_data, **kwargs):
        super(Tool, self).__init__(json_data=json_data, **kwargs)


@export
class PlacePointTool(Tool):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(PlacePointTool, self).__init__(*args, type='annotatePoint', **kwargs)


@export
class PlaceLineTool(Tool):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(PlaceLineTool, self).__init__(*args, type='annotateLine', **kwargs)


@export
class PlaceBoundingBoxTool(Tool):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(PlaceBoundingBoxTool, self).__init__(*args, type='annotateBoundingBox', **kwargs)


@export
class PlaceEllipsoidTool(Tool):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(PlaceEllipsoidTool, self).__init__(*args, type='annotateSphere', **kwargs)


tool_types = {
    'annotatePoint': PlacePointTool,
    'annotateLine': PlaceLineTool,
    'annotateBoundingBox': PlaceBoundingBoxTool,
    'annotateSphere': PlaceEllipsoidTool,
}


@export
def tool(json_data, _readonly=False):
    if isinstance(json_data, Tool):
        return json_data
    if isinstance(json_data, six.string_types):
        json_data = {'type': json_data}
    if not isinstance(json_data, dict):
        raise TypeError

    type_name = json_data.get('type')
    tool_type = tool_types.get(type_name)
    if tool_type is None: raise ValueError
    return tool_type(json_data, _readonly=_readonly)


tool.supports_readonly = True


@export
class Layer(JsonObjectWrapper):
    __slots__ = ()
    type = wrapped_property('type', optional(text_type))
    layer_dimensions = layerDimensions = wrapped_property('localDimensions', CoordinateSpace)
    layer_position = layerPosition = wrapped_property('localPosition',
                                                      optional(array_wrapper(np.float32)))
    tab = wrapped_property('tab', optional(text_type))
    pick = wrapped_property('pick', optional(bool))
    tool = wrapped_property('tool', optional(tool))

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        c.layer_position = interpolate_linear_optional_vectors(a.layer_position, b.layer_position,
                                                               t)
        return c


@export
class PointAnnotationLayer(Layer):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(PointAnnotationLayer, self).__init__(*args, type='pointAnnotation', **kwargs)

    points = wrapped_property('points', typed_list(array_wrapper(np.float32, 3)))

@export
class CoordinateSpaceTransform(JsonObjectWrapper):
    __slots__ = ()

    output_dimensions = outputDimensions = wrapped_property('outputDimensions', CoordinateSpace)
    input_dimensions = inputDimensions = wrapped_property('inputDimensions', optional(CoordinateSpace))
    source_rank = sourceRank = wrapped_property('sourceRank', optional(int))
    matrix = wrapped_property('matrix', optional(array_wrapper(np.float64)))

def data_source_url(x):
    if isinstance(x, (local_volume.LocalVolume, skeleton.SkeletonSource)):
        return x
    if not isinstance(x, six.string_types):
        raise TypeError
    return text_type(x)

@export
class LayerDataSubsource(JsonObjectWrapper):
    __slots__ = ()
    supports_validation = True

    def __init__(self, json_data=None, *args, **kwargs):
        if isinstance(json_data, bool):
            json_data = {'enabled': json_data}
        super(LayerDataSubsource, self).__init__(json_data, *args, **kwargs)

    enabled = wrapped_property('enabled', optional(bool))

@export
class LayerDataSource(JsonObjectWrapper):
    __slots__ = ()

    def __init__(self, json_data=None, *args, **kwargs):
        if (isinstance(json_data, six.string_types) or
            isinstance(json_data, (local_volume.LocalVolume, skeleton.SkeletonSource))):
            json_data = {'url': json_data}
        super(LayerDataSource, self).__init__(json_data, *args, **kwargs)

    url = wrapped_property('url', data_source_url)
    transform = wrapped_property('transform', optional(CoordinateSpaceTransform))
    subsources = wrapped_property('subsources', typed_string_map(LayerDataSubsource))
    enable_default_subsources = enableDefaultSubsources = wrapped_property('enableDefaultSubsources', optional(bool, True))

@export
class LayerDataSources(typed_list(LayerDataSource, validator=LayerDataSource)):
    __slots__ = ()

    def __init__(self, json_data=None, **kwargs):
        if isinstance(json_data, (LayerDataSource, six.string_types, local_volume.LocalVolume,
                                  skeleton.SkeletonSource, dict)):
            json_data = [json_data]
        elif isinstance(json_data, LayerDataSources):
            json_data = json_data.to_json()
        super(LayerDataSources, self).__init__(json_data, **kwargs)

class _AnnotationLayerOptions(object):
    __slots__ = ()
    annotation_color = annotationColor = wrapped_property('annotationColor', optional(text_type))


ShaderControls = typed_string_map((six.text_type, numbers.Number))


@export
class ImageLayer(Layer, _AnnotationLayerOptions):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(ImageLayer, self).__init__(*args, type='image', **kwargs)

    source = wrapped_property('source', LayerDataSources)
    shader = wrapped_property('shader', text_type)
    shader_controls = shaderControls = wrapped_property('shaderControls', ShaderControls)
    opacity = wrapped_property('opacity', optional(float, 0.5))
    blend = wrapped_property('blend', optional(str))
    cross_section_render_scale = crossSectionRenderScale = wrapped_property(
        'crossSectionRenderScale', optional(float, 1))

    @staticmethod
    def interpolate(a, b, t):
        c = Layer.interpolate(a, b, t)
        c.opacity = interpolate_linear(a.opacity, b.opacity, t)
        return c


def uint64_equivalence_map(obj, _readonly=False):
    if isinstance(obj, EquivalenceMap):
        return obj
    if obj is not None:
        obj = [[int(v) for v in group] for group in obj]
    return EquivalenceMap(obj, _readonly=_readonly)


@export
class SkeletonRenderingOptions(JsonObjectWrapper):
    __slots__ = ()

    shader = wrapped_property('shader', optional(text_type))
    shader_controls = shaderControls = wrapped_property('shaderControls', ShaderControls)
    mode2d = wrapped_property('mode2d', optional(text_type))
    line_width2d = lineWidth2d = wrapped_property('lineWidth2d', optional(float, 2))
    mode3d = wrapped_property('mode3d', optional(text_type))
    line_width3d = lineWidth3d = wrapped_property('lineWidth3d', optional(float, 1))


@export
class SegmentationLayer(Layer, _AnnotationLayerOptions):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(SegmentationLayer, self).__init__(*args, type='segmentation', **kwargs)

    source = wrapped_property('source', LayerDataSources)
    segments = wrapped_property('segments', typed_set(np.uint64))
    equivalences = wrapped_property('equivalences', uint64_equivalence_map)
    hide_segment_zero = hideSegmentZero = wrapped_property('hideSegmentZero', optional(bool, True))
    selected_alpha = selectedAlpha = wrapped_property('selectedAlpha', optional(float, 0.5))
    not_selected_alpha = notSelectedAlpha = wrapped_property('notSelectedAlpha', optional(float, 0))
    object_alpha = objectAlpha = wrapped_property('objectAlpha', optional(float, 1.0))
    saturation = wrapped_property('saturation', optional(float, 1.0))
    ignore_null_visible_set = ignoreNullVisibleSet = wrapped_property('ignoreNullVisibleSet', optional(bool, True))
    skeleton_rendering = skeletonRendering = wrapped_property('skeletonRendering', SkeletonRenderingOptions)

    @property
    def skeleton_shader(self):
        return self.skeleton_rendering.shader

    @skeleton_shader.setter
    def skeleton_shader(self, shader):
        self.skeleton_rendering.shader = shader

    skeletonShader = skeleton_shader

    color_seed = colorSeed = wrapped_property('colorSeed', optional(int, 0))
    cross_section_render_scale = crossSectionRenderScale = wrapped_property(
        'crossSectionRenderScale', optional(float, 1))
    mesh_render_scale = meshRenderScale = wrapped_property('meshRenderScale', optional(float, 10))
    mesh_silhouette_rendering = meshSilhouetteRendering = wrapped_property('meshSilhouetteRendering', optional(float, 0))
    segment_query = segmentQuery = wrapped_property('segmentQuery', optional(text_type))
    segment_colors = segmentColors = wrapped_property(
        'segmentColors', typed_map(key_type=np.uint64, value_type=text_type))

    @property
    def segment_html_color_dict(self):
        """Returns a dictionary whose keys are segments and values are the 6-digit hex
        strings representing the colors of those segments given the current
        color seed
        """
        d = {}
        for segment in self.segments:
            hex_string = segment_colors.hex_string_from_segment_id(color_seed=self.color_seed,
                                                                   segment_id=segment)
            d[segment] = hex_string
        return d

    linked_segmentation_layer = linkedSegmentationLayer = wrapped_property('linkedSegmentationLayer', optional(text_type))

    @staticmethod
    def interpolate(a, b, t):
        c = Layer.interpolate(a, b, t)
        for k in ['selected_alpha', 'not_selected_alpha', 'object_alpha']:
            setattr(c, k, interpolate_linear(getattr(a, k), getattr(b, k), t))
        return c

@export
class SingleMeshLayer(Layer):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(SingleMeshLayer, self).__init__(*args, type='mesh', **kwargs)

    source = wrapped_property('source', LayerDataSources)
    vertex_attribute_sources = vertexAttributeSources = wrapped_property(
        'vertexAttributeSources', optional(typed_list(text_type)))
    shader = wrapped_property('shader', text_type)
    vertex_attribute_names = vertexAttributeNames = wrapped_property('vertexAttributeNames',
                                                                     optional(
                                                                         typed_list(
                                                                             optional(text_type))))


class AnnotationBase(JsonObjectWrapper):
    __slots__ = ()

    id = wrapped_property('id', optional(text_type))  # pylint: disable=invalid-name
    type = wrapped_property('type', text_type)
    description = wrapped_property('description', optional(text_type))
    segments = wrapped_property('segments', optional(typed_list(typed_list(np.uint64))))
    props = wrapped_property('props', optional(typed_list(number_or_string)))


@export
class PointAnnotation(AnnotationBase):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(PointAnnotation, self).__init__(*args, type='point', **kwargs)

    point = wrapped_property('point', array_wrapper(np.float32))


@export
class LineAnnotation(AnnotationBase):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(LineAnnotation, self).__init__(*args, type='line', **kwargs)

    point_a = pointA = wrapped_property('pointA', array_wrapper(np.float32))
    point_b = pointB = wrapped_property('pointB', array_wrapper(np.float32))


@export
class AxisAlignedBoundingBoxAnnotation(AnnotationBase):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(AxisAlignedBoundingBoxAnnotation, self).__init__(
            *args, type='axis_aligned_bounding_box', **kwargs)

    point_a = pointA = wrapped_property('pointA', array_wrapper(np.float32))
    point_b = pointB = wrapped_property('pointB', array_wrapper(np.float32))


@export
class EllipsoidAnnotation(AnnotationBase):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(EllipsoidAnnotation, self).__init__(*args, type='ellipsoid', **kwargs)

    center = wrapped_property('center', array_wrapper(np.float32))
    radii = wrapped_property('radii', array_wrapper(np.float32))


annotation_types = {
    'point': PointAnnotation,
    'line': LineAnnotation,
    'axis_aligned_bounding_box': AxisAlignedBoundingBoxAnnotation,
    'ellipsoid': EllipsoidAnnotation,
}


def annotation(obj, _readonly=False):
    if isinstance(obj, AnnotationBase):
        obj = obj.to_json()
    elif not isinstance(obj, dict):
        raise TypeError
    t = obj.get('type')
    return annotation_types[t](obj, _readonly=_readonly)


annotation.supports_readonly = True

@export
class AnnotationPropertySpec(JsonObjectWrapper):
    __slots__ = ()
    id = wrapped_property('id', text_type)
    type = wrapped_property('type', text_type)
    description = wrapped_property('description', optional(text_type))
    default = wrapped_property('default', optional(number_or_string))

@export
class AnnotationLayer(Layer, _AnnotationLayerOptions):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(AnnotationLayer, self).__init__(*args, type='annotation', **kwargs)

    source = wrapped_property('source', LayerDataSources)
    annotations = wrapped_property('annotations', typed_list(annotation))
    annotation_properties = annotationProperties = wrapped_property('annotationProperties', typed_list(AnnotationPropertySpec))
    annotation_relationships = annotationRelationships = wrapped_property('annotationRelationships', typed_list(text_type))
    linked_segmentation_layer = linkedSegmentationLayer = wrapped_property('linkedSegmentationLayer', typed_string_map(text_type))
    filter_by_segmentation = filterBySegmentation = wrapped_property('filterBySegmentation', typed_list(text_type))
    ignore_null_segment_filter = ignoreNullSegmentFilter = wrapped_property('ignoreNullSegmentFilter', optional(bool, True))
    shader = wrapped_property('shader', text_type)
    shader_controls = shaderControls = wrapped_property('shaderControls', ShaderControls)

    @staticmethod
    def interpolate(a, b, t):
        del b
        del t
        return a

@export
class LocalAnnotationLayer(AnnotationLayer):
    __slots__ = ()
    def __init__(self, dimensions, *args, **kwargs):
        super(LocalAnnotationLayer, self).__init__(
            *args,
            source=LayerDataSource(
                url='local://annotations',
                transform=CoordinateSpaceTransform(outputDimensions=dimensions)),
            **kwargs)

layer_types = {
    'image': ImageLayer,
    'segmentation': SegmentationLayer,
    'pointAnnotation': PointAnnotationLayer,
    'annotation': AnnotationLayer,
    'mesh': SingleMeshLayer,
}


def make_layer(json_data, _readonly=False):
    if isinstance(json_data, Layer):
        return json_data

    if isinstance(json_data, local_volume.LocalVolume):
        json_data = dict(type=json_data.volume_type, source=json_data)

    if not isinstance(json_data, dict):
        raise TypeError

    type_name = json_data.get('type')
    layer_type = layer_types.get(type_name)
    if layer_type is not None:
        return layer_type(json_data, _readonly=_readonly)
    else:
        raise ValueError


@export
class ManagedLayer(JsonObjectWrapper):
    __slots__ = ('name', 'layer')

    def __init__(self, name, layer=None, _readonly=False, **kwargs):
        if isinstance(name, ManagedLayer):
            if layer is not None or kwargs:
                raise ValueError
            layer = name.to_json()
            name = name.name

        object.__setattr__(self, 'name', name)

        if isinstance(layer, Layer):
            json_data = collections.OrderedDict()
        elif isinstance(layer, local_volume.LocalVolume):
            json_data = collections.OrderedDict()
            layer = make_layer(layer, _readonly=_readonly)
        else:
            if layer is None:
                json_data = collections.OrderedDict()
            else:
                json_data = layer
            layer = make_layer(json_data, _readonly=_readonly)

        object.__setattr__(self, 'layer', layer)
        super(ManagedLayer, self).__init__(json_data, _readonly=_readonly, **kwargs)

    visible = wrapped_property('visible', optional(bool, True))

    def __getattr__(self, key):
        return getattr(self.layer, key)

    def __setattr__(self, key, value):
        if self._readonly:
            raise AttributeError
        if key in ['name', 'visible', 'layer']:
            object.__setattr__(self, key, value)
        else:
            return setattr(self.layer, key, value)

    def __repr__(self):
        return u'ManagedLayer(%s,%s)' % (encode_json_for_repr(self.name),
                                         encode_json_for_repr(self.to_json()))

    def to_json(self):
        r = self.layer.to_json()
        r['name'] = self.name
        visible = self.visible
        if visible:
            r.pop('visible', None)
        else:
            r['visible'] = False
        return r

    def __deepcopy__(self, memo):
        return ManagedLayer(self.name, copy.deepcopy(self.to_json(), memo))


@export
class Layers(object):
    __slots__ = ('_layers', '_readonly')
    supports_readonly = True

    def __init__(self, json_data, _readonly=False):
        if json_data is None:
            json_data = collections.OrderedDict()
        self._layers = []
        self._readonly = _readonly
        if isinstance(json_data, collections_abc.Mapping):
            for k, v in six.iteritems(json_data):
                self._layers.append(ManagedLayer(k, v, _readonly=_readonly))
        else:
            # layers property can also be an array in JSON now. each layer has a name property
            for layer in json_data:
                if isinstance(layer, ManagedLayer):
                    self._layers.append(ManagedLayer(layer.name, layer, _readonly=_readonly))
                elif isinstance(layer, dict):
                    self._layers.append(ManagedLayer(text_type(layer['name']), layer, _readonly=_readonly))
                else:
                    raise TypeError

    def index(self, k):
        for i, u in enumerate(self._layers):
            if u.name == k:
                return i
        return -1

    def __contains__(self, k):
        return self.index(k) != -1

    def __getitem__(self, k):
        """Indexes into the list of layers by index, slice, or layer name."""
        if isinstance(k, six.string_types):
            return self._layers[self.index(k)]
        return self._layers[k]

    def __setitem__(self, k, v):
        if self._readonly:
            raise AttributeError
        if isinstance(k, six.string_types):
            i = self.index(k)
            if isinstance(v, Layer):
                v = ManagedLayer(k, v)
            elif not isinstance(v, ManagedLayer):
                raise TypeError
            if i == -1:
                self._layers.append(v)
            else:
                self._layers[i] = v
        else:
            if isinstance(k, slice):
                values = []
                for x in v:
                    if not isinstance(v, ManagedLayer):
                        raise TypeError
                    values.append(x)
                self._layers[k] = values
            else:
                if not isinstance(v, ManagedLayer):
                    raise TypeError
                self._layers[k] = v

    def clear(self):
        """Clears the list of layers."""
        del self[:]

    def __delitem__(self, k):
        """Deletes a layer by index, slice, or name."""
        if self._readonly:
            raise AttributeError
        if isinstance(k, six.string_types):
            k = self.index(k)
        del self._layers[k]

    def append(self, *args, **kwargs):
        """Appends a ManagedLayer to the list of layers."""
        if self._readonly:
            raise AttributeError
        if len(args) == 1 and not kwargs and isinstance(args[0], ManagedLayer):
            layer = args[0]
        else:
            layer = ManagedLayer(*args, **kwargs)
        self._layers.append(layer)

    def extend(self, elements):
        for element in elements:
            self.append(element)

    def __len__(self):
        """Returns the number of layers in the list."""
        return len(self._layers)

    def __iter__(self):
        return iter(self._layers)

    def to_json(self):
        r = []
        for x in self._layers:
            r.append(x.to_json())
        return r

    def __repr__(self):
        return repr(self._layers)

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        for layer in c:
            index = b.index(layer.name)
            if index == -1:
                continue
            other_layer = b[index]
            if type(other_layer.layer) is not type(layer.layer):  # pylint: disable=unidiomatic-typecheck
                continue
            layer.layer = type(layer.layer).interpolate(layer.layer, other_layer.layer, t)
        return c



def navigation_link_type(x):
    x = six.text_type(x)
    x = x.lower()
    if x not in [u'linked', u'unlinked', u'relative']:
        raise ValueError('Invalid navigation link type: %r' % x)
    return x


def make_linked_navigation_type(value_type, interpolate_function=None):
    if interpolate_function is None:
        interpolate_function = value_type.interpolate

    class LinkedType(JsonObjectWrapper):
        __slots__ = ()
        link = wrapped_property('link', optional(navigation_link_type, u'linked'))
        value = wrapped_property('value', optional(value_type))

        @staticmethod
        def interpolate(a, b, t):
            c = copy.deepcopy(a)
            c.link = a.link
            if a.link == b.link and a.link != 'linked':
                c.value = interpolate_function(a, b, t)
                return c
            return c

    return LinkedType


@export
class LinkedPosition(make_linked_navigation_type(array_wrapper(np.float32), interpolate_linear_optional_vectors)):
    __slots__ = ()


@export
class LinkedZoomFactor(make_linked_navigation_type(float, interpolate_zoom)):
    __slots__ = ()


@export
class LinkedDepthRange(make_linked_navigation_type(float, interpolate_zoom)):
    __slots__ = ()


@export
class LinkedOrientationState(make_linked_navigation_type(array_wrapper(np.float32, 4), quaternion_slerp)):
    __slots__ = ()



@export
class CrossSection(JsonObjectWrapper):
    __slots__ = ()
    supports_validation = True
    width = wrapped_property('width', optional(int, 1000))
    height = wrapped_property('height', optional(int, 1000))
    position = wrapped_property('position', LinkedPosition)
    orientation = wrapped_property('orientation', LinkedOrientationState)
    scale = wrapped_property('scale', LinkedZoomFactor)

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        c.width = interpolate_linear(a.width, b.width, t)
        c.height = interpolate_linear(a.height, b.height, t)
        c.position = LinkedPosition.interpolate(a.position, b.position, t)
        c.orientation = LinkedOrientationState.interpolate(a.orientation, b.orientation, t)
        c.scale = LinkedZoomFactor.interpolate(a.scale, b.scale, t)
        return c


@export
class CrossSectionMap(typed_string_map(CrossSection)):
    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        for k in a:
            if k in b:
                c[k] = CrossSection.interpolate(a[k], b[k], t)
        return c


@export
class DataPanelLayout(JsonObjectWrapper):
    __slots__ = ()
    type = wrapped_property('type', text_type)
    cross_sections = crossSections = wrapped_property('crossSections',
                                                      CrossSectionMap)
    orthographic_projection = orthographicProjection = wrapped_property(
        'orthographicProjection', optional(bool, False))

    def __init__(self, json_data=None, _readonly=False, **kwargs):
        if isinstance(json_data, six.string_types):
            json_data = {'type': six.text_type(json_data)}
        super(DataPanelLayout, self).__init__(json_data, _readonly=_readonly, **kwargs)

    def to_json(self):
        if len(self.cross_sections) == 0 and not self.orthographic_projection:
            return self.type
        return super(DataPanelLayout, self).to_json()

    @staticmethod
    def interpolate(a, b, t):
        if a.type != b.type or len(a.cross_sections) == 0:
            return a
        c = copy.deepcopy(a)
        c.cross_sections = CrossSectionMap.interpolate(a.cross_sections, b.cross_sections, t)
        return c


def data_panel_layout_wrapper(default_value='xy'):
    def wrapper(x, _readonly=False):
        if x is None:
            x = default_value
        if isinstance(x, six.string_types):
            x = {'type': six.text_type(x)}
        return DataPanelLayout(x, _readonly=_readonly)

    wrapper.supports_readonly = True
    return wrapper


data_panel_layout_types = frozenset(['xy', 'yz', 'xz', 'xy-3d', 'yz-3d', 'xz-3d', '4panel', '3d'])


def layout_specification(x, _readonly=False):
    if x is None:
        x = '4panel'
    if isinstance(x, six.string_types):
        x = {'type': six.text_type(x)}
    if isinstance(x, (StackLayout, LayerGroupViewer, DataPanelLayout)):
        return type(x)(x.to_json(), _readonly=_readonly)
    if not isinstance(x, dict):
        raise ValueError
    layout_type = layout_types.get(x.get('type'))
    if layout_type is None:
        raise ValueError
    return layout_type(x, _readonly=_readonly)


layout_specification.supports_readonly = True


@export
class StackLayout(JsonObjectWrapper):
    __slots__ = ()
    type = wrapped_property('type', text_type)
    children = wrapped_property('children', typed_list(layout_specification))

    def __getitem__(self, key):
        return self.children[key]

    def __len__(self):
        return len(self.children)

    def __setitem__(self, key, value):
        self.children[key] = value

    def __delitem__(self, key):
        del self.children[key]

    def __iter__(self):
        return iter(self.children)

    @staticmethod
    def interpolate(a, b, t):
        if a.type != b.type or len(a.children) != len(b.children):
            return a
        c = copy.deepcopy(a)
        c.children = [
            interpolate_layout(a_child, b_child, t)
            for a_child, b_child in zip(a.children, b.children)
        ]
        return c


@export
def row_layout(children):
    return StackLayout(type='row', children=children)


@export
def column_layout(children):
    return StackLayout(type='column', children=children)


def interpolate_layout(a, b, t):
    if type(a) is not type(b):
        return a
    return type(a).interpolate(a, b, t)


@export
class LayerGroupViewer(JsonObjectWrapper):
    __slots__ = ()
    type = wrapped_property('type', text_type)
    layers = wrapped_property('layers', typed_list(text_type))
    layout = wrapped_property('layout', data_panel_layout_wrapper('xy'))
    position = wrapped_property('position', LinkedPosition)
    cross_section_orientation = crossSectionOrientation = wrapped_property(
        'crossSectionOrientation', LinkedOrientationState)
    cross_section_scale = crossSectionScale = wrapped_property('crossSectionZoom', LinkedZoomFactor)
    cross_section_depth = crossSectionDepth = wrapped_property('crossSectionDepth', LinkedDepthRange)
    projection_orientation = projectionOrientation = wrapped_property(
        'projectionOrientation', LinkedOrientationState)
    projection_scale = projectionScale = wrapped_property('projectionScale', LinkedZoomFactor)
    projection_depth = projectionDepth = wrapped_property('projectionDepth', LinkedDepthRange)

    def __init__(self, *args, **kwargs):
        super(LayerGroupViewer, self).__init__(*args, **kwargs)
        self.type = 'viewer'

    def __repr__(self):
        j = self.to_json()
        j.pop('type', None)
        return u'%s(%s)' % (type(self).__name__, encode_json_for_repr(j))

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        for k in ('layout', 'position', 'cross_section_orientation', 'cross_section_zoom',
                  'perspective_orientation', 'perspective_zoom'):
            a_attr = getattr(a, k)
            b_attr = getattr(b, k)
            setattr(c, k, type(a_attr).interpolate(a_attr, b_attr, t))
        return c


layout_types = {
    'row': StackLayout,
    'column': StackLayout,
    'viewer': LayerGroupViewer,
}

def add_data_panel_layout_types():
    for k in data_panel_layout_types:
        layout_types[k] = DataPanelLayout
add_data_panel_layout_types()


@export
class SelectedLayerState(JsonObjectWrapper):
    visible = wrapped_property('visible', optional(bool, False))
    size = wrapped_property('size', optional(int))
    layer = wrapped_property('layer', optional(text_type))


@export
class StatisticsDisplayState(JsonObjectWrapper):
    visible = wrapped_property('visible', optional(bool, False))
    size = wrapped_property('size', optional(int))


@export
class ViewerState(JsonObjectWrapper):
    __slots__ = ()
    dimensions = wrapped_property('dimensions', CoordinateSpace)
    relative_display_scales = relativeDisplayScales = wrapped_property('relativeDisplayScales', optional(typed_string_map(float)))
    display_dimensions = displayDimensions = wrapped_property('displayDimensions', optional(typed_list(text_type)))
    position = voxel_coordinates = wrapped_property('position', optional(array_wrapper(np.float32)))
    cross_section_orientation = crossSectionOrientation = wrapped_property(
        'crossSectionOrientation', optional(array_wrapper(np.float32, 4)))
    cross_section_scale = crossSectionScale = wrapped_property('crossSectionScale', optional(float))
    cross_section_depth = crossSectionDepth = wrapped_property('crossSectionDepth', optional(float))
    projection_scale = projectionScale = wrapped_property('projectionScale', optional(float))
    projection_depth = projectionDepth = wrapped_property('projectionDepth', optional(float))
    projection_orientation = projectionOrientation = perspectiveOrientation = perspective_orientation = wrapped_property(
        'projectionOrientation', optional(array_wrapper(np.float32, 4)))
    show_slices = showSlices = wrapped_property('showSlices', optional(bool, True))
    show_axis_lines = showAxisLines = wrapped_property('showAxisLines', optional(bool, True))
    show_scale_bar = showScaleBar = wrapped_property('showScaleBar', optional(bool, True))
    show_default_annotations = showDefaultAnnotations = wrapped_property('showDefaultAnnotations', optional(bool, True))
    gpu_memory_limit = gpuMemoryLimit = wrapped_property('gpuMemoryLimit', optional(int))
    system_memory_limit = systemMemoryLimit = wrapped_property('systemMemoryLimit', optional(int))
    concurrent_downloads = concurrentDownloads = wrapped_property('concurrentDownloads', optional(int))
    prefetch = wrapped_property('prefetch', optional(bool, True))
    layers = wrapped_property('layers', Layers)
    layout = wrapped_property('layout', layout_specification)
    cross_section_background_color = crossSectionBackgroundColor = wrapped_property(
        'crossSectionBackgroundColor', optional(text_type))
    projection_background_color = projectionBackgroundColor = wrapped_property(
        'projectionBackgroundColor', optional(text_type))
    selected_layer = selectedLayer = wrapped_property('selectedLayer', SelectedLayerState)
    statistics = wrapped_property('statistics', StatisticsDisplayState)
    partial_viewport = partialViewport = wrapped_property(
        'partialViewport',
        optional(array_wrapper(np.float64, 4), np.array([0, 0, 1, 1], dtype=np.float64)))

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        c.position = interpolate_linear_optional_vectors(a.position, b.position, t)
        c.projection_scale = interpolate_zoom(a.projection_scale, b.projection_scale, t)
        c.projection_orientation = quaternion_slerp(a.projection_orientation,
                                                     b.projection_orientation, t)
        c.cross_section_scale = interpolate_zoom(a.cross_section_scale, b.cross_section_scale, t)
        c.cross_section_orientation = quaternion_slerp(a.cross_section_orientation,
                                                       b.cross_section_orientation, t)
        c.layers = Layers.interpolate(a.layers, b.layers, t)
        c.layout = interpolate_layout(a.layout, b.layout, t)
        return c



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

def hash_function(state,value):
    """ Python implementation of hashCombine() function
    in src/neuroglancer/gpu_hash/hash_function.ts,
    a modified murmur hash
    """
    k1 = 0xcc9e2d51
    k2 = 0x1b873593
    state = state & 0xffffffff
    value = (value * k1) & 0xffffffff
    value = ((value << 15) | value >> 17) & 0xffffffff
    value = (value * k2) & 0xffffffff
    state = (state ^ value) & 0xffffffff
    state = (( state << 13) | state >> 19) & 0xffffffff
    state = (( state * 5) + 0xe6546b64) & 0xffffffff
    return state

def hsv_to_rgb(h,s,v):
    """ Convert H,S,V values to RGB values.
    Python implementation of hsvToRgb in src/neuroglancer/util/colorspace.ts """
    h*=6
    hue_index = math.floor(h)
    remainder = h - hue_index
    val1 = v*(1-s)
    val2 = v*(1-(s*remainder))
    val3 = v*(1-(s*(1-remainder)))
    hue_remainder = hue_index % 6
    if hue_remainder == 0:
        return (v,val3,val1)
    elif hue_remainder == 1:
        return (val2,v,val1)
    elif hue_remainder == 2:
        return (val1,v,val3)
    elif hue_remainder == 3:
        return (val1,val2,v)
    elif hue_remainder == 4:
        return (val3,val1,v)
    elif hue_remainder == 5: 
        return (v,val1,val2)   

def pack_color(rgb_vec):
    """ Returns an integer formed
    by concatenating the channels of the input color vector.
    Python implementation of packColor in src/neuroglancer/util/color.ts
    """
    result = 0
    for i in range(len(rgb_vec)):
        result = ((result << 8) & 0xffffffff) + min(255,max(0,round(rgb_vec[i]*255)))
    return result

def hex_string_from_segment_id(color_seed,segment_id):
    """ Return the hex color string for a segment
    given a color seed and the segment id """
    segment_id = int(segment_id) # necessary since segment_id is 64 bit originally 
    result = hash_function(state=color_seed,value=segment_id)
    newvalue = segment_id >> 32
    result2 = hash_function(state=result,value=newvalue)
    c0 = (result2 & 0xFF) / 255.
    c1 = ((result2 >> 8) & 0xFF) / 255.;
    h = c0
    s =  0.5 + 0.5 * c1
    v = 1.0
    rgb=hsv_to_rgb(h,s,v)
    packed_color = pack_color(rgb_vec=rgb)
    hex_string = format(packed_color, 'x')
    """ Zero pad the hex string if less than 6 characeters """
    if len(hex_string) < 6:
        hex_string = '0'*(6-len(hex_string)) + hex_string
    hex_string = '#' + hex_string
    return hex_string


#!/usr/bin/env python
# @license
# Copyright 2018 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division

import argparse
import collections
import copy
import json
import math
import os
import uuid
import webbrowser

import numpy as np

import neuroglancer
import neuroglancer.cli
import neuroglancer.url_state
from neuroglancer.json_utils import json_encoder_default

def get_segmentation_layer(layers):
    for layer in layers:
        if isinstance(layer.layer, neuroglancer.SegmentationLayer):
            return layer

def _full_count_for_level(level):
    return (2**level)**3

class BlockMask(object):
    def __init__(self, max_level=3):
        # self.blocks[level_i][position] specifies the number of base elements contained within the block
        self.blocks = [dict() for _ in range(max_level+1)]

    def _remove_children(self, level, position):
        position = tuple(position)
        if level == 0:
            return
        for offset in np.ndindex((2,) * 3):
            self._remove(level - 1, tuple(x * 2 + o for x, o in zip(position, offset)))

    def _remove(self, level, position):
        position = tuple(position)
        old_count = self.blocks[level].pop(position, 0)
        if old_count == 0 or old_count == _full_count_for_level(level):
            return old_count
        self._remove_children(level, position)
        return old_count

    def _contains(self, level, position):
        position = tuple(position)
        blocks = self.blocks
        while True:
            count = blocks[level].get(position, 0)
            if count == _full_count_for_level(level):
                return level, position
            position = tuple(x // 2 for x in position)
            level += 1
            if level >= len(blocks):
                return None, None

    def _add_children(self, level, position, excluded_child_position, excluded_child_count):
        blocks = self.blocks
        full_count_for_child = _full_count_for_level(level-1)
        for offset in np.ndindex((2,) * 3):
            child_position = tuple(x * 2 + o for x, o in zip(position, offset))
            count = full_count_for_child
            if child_position == excluded_child_position:
                count -= excluded_child_count
            if count != 0:
                blocks[level-1][child_position] = count

    def _add_children_along_path(self, start_level, end_level, start_position):
        excluded_count = _full_count_for_level(start_level)
        while start_level < end_level:
            parent_position = tuple(x // 2 for x in start_position)
            start_level += 1
            self._add_children(start_level, parent_position, start_position, excluded_count)
            start_position = parent_position


    def add(self, level, position):
        if self._contains(level, position)[0] is not None:
            return
        position = tuple(position)
        old_count = self.blocks[level].get(position, 0)
        self._adjust_count(level, position, _full_count_for_level(level) - old_count)

    def remove(self, level, position):
        position = tuple(position)
        old_count = self.blocks[level].get(position, 0)
        if old_count == 0:
            old_level, position_in_old_level = self._contains(level, position)
            if old_level is None:
                return
            if old_level != level:
                self._adjust_count(old_level, position_in_old_level, -_full_count_for_level(level))
                self._add_children_along_path(level, old_level, position)
                return
        if old_count != _full_count_for_level(level):
            self._remove_children(level, position)
        self._adjust_count(level, position, -old_count)

    def _adjust_count(self, level, position, amount):
        if amount == 0:
            return
        old_count = self.blocks[level].get(position, 0)
        new_count = old_count + amount
        if new_count == 0:
            del self.blocks[level][position]
        else:
            self.blocks[level][position] = new_count

        if level > 0 and new_count == _full_count_for_level(level):
            self._remove_children(level, position)

        if level + 1 < len(self.blocks):
            self._adjust_count(level + 1, tuple(x // 2 for x in position), amount)

def make_block_mask(annotations, block_size, max_level=3):
    mask = BlockMask(max_level=max_level)
    for x in annotations:
        if not isinstance(x, neuroglancer.AxisAlignedBoundingBoxAnnotation):
            print('Warning: got non-box annotation: %r' % (x,))
            continue
        size = (x.point_b - x.point_a) / block_size
        if size[0] != int(size[0]) or np.any(size != size[0]):
            print('Warning: got invalid box: %r' % (x,))
            continue
        level = math.log(size[0]) / math.log(2)
        if level != int(level):
            print('Warning: got invalid box: %r' % (x,))
            continue
        level = int(level)
        eff_block_size = block_size * (2**level)
        if np.any(x.point_a % eff_block_size != 0):
            print('Warning: got invalid box: %r' % (x,))
            continue
        position = tuple(int(z) for z in x.point_a // eff_block_size)
        mask.add(level, position)
    return mask

def make_annotations_from_mask(mask, block_size):
    result = []
    for level, position_counts in enumerate(mask.blocks):
        full_count = _full_count_for_level(level)
        eff_block_size = block_size * 2**level
        for position in position_counts:
            count = position_counts[position]
            if count != full_count:
                continue
            position = np.array(position, dtype=np.int64)
            box_start = eff_block_size * position
            box_end = box_start + eff_block_size
            result.append(neuroglancer.AxisAlignedBoundingBoxAnnotation(
                point_a = box_start,
                point_b = box_end,
                id = uuid.uuid4().hex,
            ))
    return result


def normalize_block_annotations(annotations, block_size, max_level=3):
    mask = make_block_mask(annotations=annotations, block_size=block_size, max_level=max_level)
    return make_annotations_from_mask(mask=mask, block_size=block_size)


class Annotator(object):
    def __init__(self, filename):
        self.filename = filename
        self.annotation_layer_name = 'false-merges'
        self.states = []
        self.state_index = None
        self.false_merge_block_size = np.array([32, 32, 32], dtype=np.int64)
        self.false_merge_block_level = 0
        self.max_false_merge_block_levels = 3
        viewer = self.viewer = neuroglancer.Viewer()
        self.other_state_segment_ids = dict()

        viewer.actions.add('anno-next-state', lambda s: self.next_state())
        viewer.actions.add('anno-prev-state', lambda s: self.prev_state())
        viewer.actions.add('anno-save', lambda s: self.save())
        viewer.actions.add('anno-show-all', lambda s: self.set_combined_state())
        viewer.actions.add('anno-add-segments-from-state',
                           lambda s: self.add_segments_from_state(s.viewer_state))
        viewer.actions.add('anno-mark-false-merge', self.mark_false_merge)
        viewer.actions.add('anno-unmark-false-merge', lambda s: self.mark_false_merge(s, erase=True))
        viewer.actions.add('anno-decrease-block-size', self.decrease_false_merge_block_size)
        viewer.actions.add('anno-increase-block-size', self.increase_false_merge_block_size)

        with viewer.config_state.txn() as s:
            s.input_event_bindings.data_view['pageup'] = 'anno-prev-state'
            s.input_event_bindings.data_view['pagedown'] = 'anno-next-state'
            s.input_event_bindings.data_view['bracketleft'] = 'anno-decrease-block-size'
            s.input_event_bindings.data_view['bracketright'] = 'anno-increase-block-size'
            s.input_event_bindings.data_view['control+keys'] = 'anno-save'
            s.input_event_bindings.data_view['control+keya'] = 'anno-show-all'
            s.input_event_bindings.data_view['control+mousedown0'] = 'anno-mark-false-merge'
            s.input_event_bindings.data_view['control+shift+mousedown0'] = 'anno-unmark-false-merge'

        viewer.shared_state.add_changed_callback(self.on_state_changed)
        self.cur_message = None
        if not self.load():
            self.set_state_index(None)

    def increase_false_merge_block_size(self, s):
        self.false_merge_block_level = min(self.max_false_merge_block_levels, self.false_merge_block_level + 1)
        self.update_message()

    def decrease_false_merge_block_size(self, s):
        self.false_merge_block_level = max(0, self.false_merge_block_level - 1)
        self.update_message()

    def mark_false_merge(self, s, erase=False):
        voxel_coordinates = s.mouse_voxel_coordinates
        if voxel_coordinates is None:
            return
        block_size = self.false_merge_block_size
        level = self.false_merge_block_level
        full_block_size = block_size * 2**level
        block_position = voxel_coordinates // full_block_size

        with self.viewer.txn() as s:
            annotations = s.layers[self.annotation_layer_name].annotations
            mask = make_block_mask(annotations=annotations, block_size=block_size, max_level=self.max_false_merge_block_levels)
            if erase:
                mask.remove(level, block_position)
            else:
                mask.add(level, block_position)
            new_annotations = make_annotations_from_mask(mask=mask, block_size=block_size)
            s.layers[self.annotation_layer_name].annotations = new_annotations

    def on_state_changed(self):
        # Check if we should warn about segments
        self.update_message()

    def update_message(self):
        message = '[Block size: %d vx] ' % (self.false_merge_block_size[0] * 2**self.false_merge_block_level)
        if self.state_index is None:
            message += '[No state selected]'
        else:
            message += '[%d/%d] ' % (self.state_index, len(self.states))
            segments = self.get_state_segment_ids(self.viewer.state)
            warnings = []
            for segment_id in segments:
                other_state = self.other_state_segment_ids.get(segment_id)
                if other_state is not None:
                    warnings.append('Segment %d also in state %d' % (segment_id, other_state))
            if warnings:
                message += 'WARNING: ' + ', '.join(warnings)
        if message != self.cur_message:
            with self.viewer.config_state.txn() as s:
                if message is not None:
                    s.status_messages['status'] = message
                else:
                    s.status_messages.pop('status')
            self.cur_message = message

    def load(self):
        if not os.path.exists(self.filename):
            return False
        self.state_index = None

        with open(self.filename, 'r') as f:

            loaded_state = json.load(f, object_pairs_hook=collections.OrderedDict)
        self.states = [neuroglancer.ViewerState(x) for x in loaded_state['states']]
        self.set_state_index(loaded_state['state_index'])
        return True

    def set_state_index_relative(self, amount):
        if self.state_index is None:
            new_state = 0
        else:
            new_state = (self.state_index + amount + len(self.states)) % len(self.states)
        self.set_state_index(new_state)

    def next_state(self):
        self.set_state_index_relative(1)

    def prev_state(self):
        self.set_state_index_relative(-1)

    def set_state_index(self, index):
        self._grab_viewer_state()
        self.state_index = index
        if index is None:
            self.viewer.set_state(neuroglancer.ViewerState())
        else:
            new_state = copy.deepcopy(self.states[index])
            anno_layer = new_state.layers[self.annotation_layer_name]
            if anno_layer.annotation_fill_opacity == 0:
                anno_layer.annotation_fill_opacity = 0.7
            anno_layer.annotation_color = 'black'
            anno_layer.annotations = normalize_block_annotations(
                anno_layer.annotations,
                block_size=self.false_merge_block_size,
                max_level=self.max_false_merge_block_levels)
            self.viewer.set_state(new_state)
            other_ids = self.other_state_segment_ids
            other_ids.clear()
            other_ids[0] = -1
            for i, state in enumerate(self.states):
                if i == self.state_index:
                    continue
                for x in self.get_state_segment_ids(state):
                    other_ids[x] = i
        self.update_message()

    def get_duplicate_segment_ids(self):
        self._grab_viewer_state()
        other_ids = dict()
        other_ids[0] = [-1]
        for i, state in enumerate(self.states):
            for x in self.get_state_segment_ids(state):
                other_ids.setdefault(x, []).append(i)
        for segment_id in other_ids:
            state_numbers = other_ids[segment_id]
            if len(state_numbers) > 1:
                print('%d in %r' % (segment_id, state_numbers))

    def _grab_viewer_state(self):
        if self.state_index is not None:
            self.states[self.state_index] = copy.deepcopy(self.viewer.state)

    def save(self):
        self._grab_viewer_state()
        tmp_filename = self.filename + '.tmp'
        with open(tmp_filename, 'wb') as f:
            f.write(
                json.dumps(
                    dict(states=[s.to_json() for s in self.states], state_index=self.state_index),
                    default=json_encoder_default))
        os.rename(tmp_filename, self.filename)
        print('Saved state to: %s' % (self.filename, ))

    def get_state_segment_ids(self, state):
        return get_segmentation_layer(state.layers).segments

    def get_existing_segment_ids(self):
        ids = set()
        for state in self.states:
            ids.update(self.get_state_segment_ids(state))
        return ids

    def add_segments_from_state(self, base_state):
        if isinstance(base_state, basestring):
            base_state = neuroglancer.parse_url(base_state)
        elif isinstance(base_state, dict):
            base_state = neuroglancer.ViewerState(base_state)

        segment_ids = self.get_state_segment_ids(base_state)

        existing_segment_ids = self.get_existing_segment_ids()

        for segment_id in segment_ids:
            if segment_id in existing_segment_ids:
                print('Skipping redundant segment id %d' % segment_id)
                continue
            self.states.append(self.make_initial_state(segment_id, base_state))

        if self.state_index is None:
            self.next_state()

    def make_initial_state(self, segment_id, base_state):
        state = copy.deepcopy(base_state)

        segments = self.get_state_segment_ids(state)
        segments.clear()
        segments.add(segment_id)
        state.layers[self.annotation_layer_name] = neuroglancer.AnnotationLayer()

        return state

    def remove_zero_segments(self):
        for state in self.states:
            segment_ids = self.get_state_segment_ids(state)
            if 0 in segment_ids:
                segment_ids.remove(0)

    def set_combined_state(self):
        state = self.make_combined_state()
        if state is None:
            print('No states')
        else:
            self.set_state_index(None)
            self.viewer.set_state(state)

    def make_combined_state(self):
        if len(self.states) == 0:
            return None

        state = copy.deepcopy(self.states[0])
        layer = get_segmentation_layer(state.layers)
        layer.segments.clear()
        points = state.layers[self.annotation_layer_name].annotations
        for i, other_state in enumerate(self.states):
            other_segments = self.get_state_segment_ids(other_state)
            # print('%d: %r' % (i, other_segments))
            if other_segments:
                u_result = layer.equivalences.union(*other_segments)
                layer.segments.add(u_result)
            points.extend(other_state.layers[self.annotation_layer_name].annotations)
        return state

    def show(self):
        webbrowser.open_new(self.viewer.get_viewer_url())

    def get_viewer_url(self):
        return self.viewer.get_viewer_url()

    def get_sets(self):
        sets = []
        for other_state in self.states:
            other_segments = self.get_state_segment_ids(other_state)
            if other_segments:
                sets.append(sorted(other_segments))
        return sets


if __name__ == '__main__':
    ap = argparse.ArgumentParser()
    ap.add_argument('filename', type=str)
    ap.add_argument('--add-segments-from-url', type=str, nargs='*', default=[])
    ap.add_argument(
        '-n', '--no-webbrowser', action='store_true', help='Don\'t open the webbrowser.')
    ap.add_argument('--print-sets', action='store_true', help='Print the sets of supervoxels.')
    ap.add_argument(
        '--print-combined-state',
        action='store_true',
        help='Prints a neuroglancer link for the combined state.')
    ap.add_argument(
        '--print-summary',
        action='store_true',
        help='Prints a neuroglancer link for the combined state.')
    neuroglancer.cli.add_server_arguments(ap)

    args = ap.parse_args()
    neuroglancer.cli.handle_server_arguments(args)

    anno = Annotator(args.filename)
    for url in args.add_segments_from_url:
        anno.add_segments_from_state(url)

    if args.print_sets:
        print(repr(anno.get_sets()))

    if args.print_combined_state:
        print(neuroglancer.to_url(anno.make_combined_state()))

    if args.print_summary:
        print('<html>')
        print('<h1>%s</h1>' % args.filename)
        print(
            '<a href="%s">Neuroglancer</a><br/>' % neuroglancer.to_url(anno.make_combined_state()))
        print(repr(anno.get_sets()))
        print('</html>')

    else:
        print(anno.get_viewer_url())
    if not args.no_webbrowser:
        anno.show()



# coding=utf-8
# @license
# Copyright 2020 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for controlling a browser that runs Neuroglancer."""

from __future__ import absolute_import

import tempfile
import time
import threading

class Webdriver(object):

    def __init__(self,
                 viewer=None,
                 headless=True,
                 browser='chrome',
                 window_size=(1920, 1080),
                 debug=False,
                 docker=False,
                 print_logs=True):
        if viewer is None:
            from .viewer import Viewer
            viewer = Viewer()
        self.viewer = viewer
        self.headless = headless
        self.browser = browser
        self.window_size = window_size
        self.headless = headless
        self.docker = docker
        self.debug = debug
        self._logfile = None
        if browser == 'firefox':
            self._logfile = tempfile.NamedTemporaryFile(suffix='neuroglancer-geckodriver.log')
        self._init_driver()
        self._pending_logs = []
        self._pending_logs_to_print = []
        self._logs_lock = threading.Lock()
        self._closed = False
        def print_log_handler():
            while True:
                logs_to_print = self._get_logs_to_print()
                if logs_to_print:
                    print('\n'.join(x['message'] for x in logs_to_print))
                if self._closed:
                    break
                time.sleep(1)
        t = threading.Thread(target = print_log_handler)
        t.daemon = True
        t.start()

    def _init_chrome(self):
        import selenium.webdriver
        import selenium.webdriver.chrome.options
        import selenium.webdriver.common.service
        import selenium.webdriver.chrome.service

        def patched_init(self, executable_path, port=0, service_args=None, log_path=None, env=None):
            log_path = '/dev/stderr'
            self.service_args = service_args or []
            if log_path:
                self.service_args.append('--log-path=%s' % log_path)
            selenium.webdriver.common.service.Service.__init__(
                self,
                executable_path,
                port=port,
                env=env,
                log_file=None,
                start_error_message=
                "Please see https://sites.google.com/a/chromium.org/chromedriver/home")

        import selenium.webdriver.common.desired_capabilities
        try:
            # Use chromedriver_binary package if available
            import chromedriver_binary
        except ImportError:
            # Fallback to system chromedriver
            pass
        chrome_options = selenium.webdriver.chrome.options.Options()
        if self.headless:
            chrome_options.add_argument('--headless')
        if self.docker:
            # https://www.intricatecloud.io/2019/05/running-webdriverio-tests-using-headless-chrome-inside-a-container/
            chrome_options.add_argument('--no-sandbox')
            chrome_options.add_argument('--disable-gpu')
            chrome_options.add_argument('--disable-setuid-sandbox')
            chrome_options.add_argument('--disable-dev-shm-usage')
        chrome_options.add_argument('--window_size=%dx%d' %
                                    (self.window_size[0], self.window_size[1]))
        caps = selenium.webdriver.common.desired_capabilities.DesiredCapabilities.CHROME.copy()
        caps['goog:loggingPrefs'] = {'browser': 'ALL'}
        try:
            orig_init = selenium.webdriver.chrome.service.Service.__init__
            if self.debug:
                selenium.webdriver.chrome.service.Service.__init__ = patched_init
            self.driver = selenium.webdriver.Chrome(options=chrome_options, desired_capabilities=caps)
        finally:
            if self.debug:
                selenium.webdriver.chrome.service.Service.__init__ = orig_init

    def _init_firefox(self):
        import selenium.webdriver
        try:
            # Use geckodriver_autoinstaller package if available
            import geckodriver_autoinstaller
            geckodriver_autoinstaller.install()
        except (ImportError, SyntaxError):
            # Fallback to system chromedriver
            pass
        profile = selenium.webdriver.FirefoxProfile()
        profile.set_preference('devtools.console.stdout.content', True)
        self.driver = selenium.webdriver.Firefox(firefox_profile=profile,
                                                 service_log_path=self._logfile.name)

    def _init_driver(self):
        if self.browser == 'chrome':
            self._init_chrome()
        elif self.browser == 'firefox':
            self._init_firefox()
        else:
            raise ValueError('unsupported browser: %s, must be "chrome" or "firefox"' %
                             (self.browser, ))
        self.driver.get(self.viewer.get_viewer_url())

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._logfile is not None:
            self._logfile.file.close()
            self._logfile = None
        self.driver.quit()
        self._closed = True

    def _get_new_logs(self):
        if self.browser == 'chrome':
            new_logs = self.driver.get_log('browser')
        else:
            cur_offset = self._logfile.file.tell()
            new_data = self._logfile.file.read()
            # rfind may return -1, still works
            end_within_data = new_data.rfind(b'\n') + 1
            new_data = new_data[:end_within_data]
            self._logfile.file.seek(cur_offset + end_within_data)
            new_logs = []
            for msg in new_data.decode().split('\n'):
                msg = msg.strip()
                if not msg: continue
                if (not msg.startswith('console.log: ')
                        and not msg.startswith('JavaScript ')):
                    continue
                new_logs.append({'message': msg})
        self._pending_logs.extend(new_logs)
        self._pending_logs_to_print.extend(new_logs)

    def _get_logs_to_print(self):
        with self._logs_lock:
            self._get_new_logs()
            new_logs = self._pending_logs_to_print
            self._pending_logs_to_print = []
            return new_logs

    def get_log(self):
        with self._logs_lock:
            self._get_new_logs()
            new_logs = self._pending_logs
            self._pending_logs = []
            return new_logs

    def get_log_messages(self):
        return '\n'.join(x['message'] for x in self.get_log())

    def sync(self):
        """Wait until client is ready."""
        while True:
            new_state = self.viewer.screenshot().viewer_state
            # Ensure self.viewer.state has also been updated to the new state.
            # The state sent in the screenshot reply can be newer.
            if new_state == self.viewer.state:
                return new_state
            time.sleep(0.1)

    def reload_browser(self):
        """Reloads the browser (useful if it crashes/becomes unresponsive)."""
        with self._logs_lock:
            try:
                self.driver.quit()
            except:
                pass
            self._init_driver()

    @property
    def root_element(self):
        return self.driver.find_element_by_xpath('//body')

    def action_chain(self):
        import selenium.webdriver
        return selenium.webdriver.common.action_chains.ActionChains(self.driver)



#!/usr/bin/env python
# @license
# Copyright 2020 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tool for creating screenshots with Neuroglancer.

The Neuroglancer state may be specified either by a URL or by a path to a JSON
state file.

Rendering requires a web browser.  By default, a headless chromedriver is
started in the background.  It is also possible to use non-headless chromedriver
or a manually-opened browser.

There are several methods by which the screenshot image may be rendered:

1. The state can be rendered directly as a single frame by Neuroglancer.  This
   is the simplest and fastest method and works for most states.

2. If the output image size exceeds what Neuroglancer/the browser can support
   (usually about 4096x4096), tiled rendering can be used.  In this case,
   Neuroglancer will render the image as multiple tiles which are assembled
   automatically into a single image.  This is enabled automatically if the
   requested image size exceeds the specified tile dimensions.  All normal
   functionality is supported, except for the "show_slices" option whereby
   cross-section panels are also shown in the 3-d view.  Manually-specified
   cross sections via the "cross_sections" property are supported, however.

3. If a very large number of 3-d objects are to be rendered, it may be
   impossible for Neuroglancer to render them all simultaneously due to memory
   limits.  The `--segment-shard-size` option may be specified to enable a
   special rendering mode in which subsets of the objects are rendered
   independently and then combined together into a single image.  Depth
   information is used to combine the images together.  Currently, transparent
   rendering of objects is not supported, though.  As the final image is
   produced incrementally, the state is saved in a `.npz` file, which allows
   resuming if the screenshot process is interrupted.  To avoid resuming if you
   change options, delete the `.npz` file.

Tips:

- The Neuroglancer UI controls are not shown, and in the case of multi-panel
  layouts, there is no border between panels.  In most cases it is desirable to
  capture a single-panel layout.

- The layer side panel and statistics panel, if open, will be closed for the
  screenshot.

- The specified image dimensions will be used, rather than the dimensions of
  your browser window.  This, in combination with the removal of the normal
  Neuroglancer UI controls, means that the field of view may differ somewhat.

- The axis lines and volume bounding boxes will be shown if they are enabled in
  the Neuroglancer state.  If you don't want them in the screenshot, you should
  disable them in the Neuroglancer state.  You may also use the
  `--hide-axis-lines` and `--hide-default-annotations` options.  In most cases
  it is desirable to hide the axis lines and default annotations.

- The scale bars will be shown if they are enabled in the Neuroglancer state.
  If you specify a large image size, you may want to increase the size of the
  scale bar, using the `--scale-bar-scale` option.

"""

import argparse
import collections
import contextlib
import copy
import datetime
import itertools
import numbers
import os
import threading
import time
from typing import NamedTuple, Tuple, Callable, Iterator, List, Optional

import PIL
import numpy as np

import neuroglancer
import neuroglancer.cli
import neuroglancer.webdriver


def _get_total_segments(state):
    num_segments = 0
    for layer in state.layers:
        if not isinstance(layer.layer, neuroglancer.SegmentationLayer):
            continue
        num_segments += len(layer.segments)
    return num_segments


def _should_shard_segments(state, segment_shard_size):
    return _get_total_segments(state) > segment_shard_size


def _calculate_num_shards(state, segment_shard_size):
    total_segments = _get_total_segments(state)
    return -(-total_segments // segment_shard_size)


def _get_sharded_states(state, segment_shard_size, reverse_bits):
    if reverse_bits:
        sort_key = lambda x: int('{:064b}'.format(x)[::-1], 2)
    else:
        sort_key = None
    num_shards = _calculate_num_shards(state, segment_shard_size)
    for shard_i in range(num_shards):
        new_state = copy.deepcopy(state)
        cum_retained = 0
        cum_skipped = segment_shard_size * shard_i
        for i, layer in enumerate(new_state.layers):
            if not isinstance(layer.layer, neuroglancer.SegmentationLayer):
                continue
            segments = sorted(layer.segments, key=sort_key)
            num_to_skip = min(cum_skipped, len(segments))
            segments = segments[num_to_skip:]
            cum_skipped += num_to_skip
            num_to_retain = min(segment_shard_size - cum_retained, len(segments))
            cum_retained += num_to_retain
            layer.segments = set(segments[:num_to_retain])
        yield new_state


class TileGenerator:
    def __init__(self, shape, tile_shape):
        self.tile_shape = tuple(tile_shape)
        self.shape = tuple(shape)
        self.tile_grid_shape = tuple(-(-self.shape[i] // self.tile_shape[i]) for i in range(2))
        self.tile_shape = tuple(-(-self.shape[i] // self.tile_grid_shape[i]) for i in range(2))
        self.num_tiles = self.tile_grid_shape[0] * self.tile_grid_shape[1]

    def get_tile_states(self, state):
        for tile_y in range(self.tile_grid_shape[1]):
            for tile_x in range(self.tile_grid_shape[0]):
                x_offset = tile_x * self.tile_shape[0]
                y_offset = tile_y * self.tile_shape[1]
                tile_width = min(self.tile_shape[0], self.shape[0] - x_offset)
                tile_height = min(self.tile_shape[1], self.shape[1] - y_offset)
                new_state = copy.deepcopy(state)
                new_state.partial_viewport = [
                    x_offset / self.shape[0], y_offset / self.shape[1], tile_width / self.shape[0],
                    tile_height / self.shape[1]
                ]
                params = {
                    'tile_x': tile_x,
                    'tile_y': tile_y,
                    'x_offset': x_offset,
                    'y_offset': y_offset,
                    'tile_width': tile_width,
                    'tile_height': tile_height,
                }
                yield params, new_state


class ShardedTileGenerator(TileGenerator):
    def __init__(self, state, segment_shard_size, reverse_bits, **kwargs):
        super(ShardedTileGenerator, self).__init__(**kwargs)
        self.state = state
        self.reverse_bits = reverse_bits
        self.total_segments = _get_total_segments(self.state)
        self.segment_shard_size = segment_shard_size
        self.num_shards = _calculate_num_shards(self.state, self.segment_shard_size)
        self.num_tiles *= self.num_shards

    def get_states(self):
        for shard_i, state in enumerate(
                _get_sharded_states(self.state,
                                    self.segment_shard_size,
                                    reverse_bits=self.reverse_bits)):
            for params, state in self.get_tile_states(state):
                params['segment_shard'] = shard_i
                yield params, state


class CaptureScreenshotRequest(NamedTuple):
    state: neuroglancer.ViewerState
    description: str
    config_callback: Callable[[neuroglancer.viewer_config_state.ConfigState], None]
    response_callback: neuroglancer.viewer_config_state.ScreenshotReply
    include_depth: bool = False


def buffered_iterator(base_iter, lock, buffer_size):
    while True:
        with lock:
            buffered_items = list(itertools.islice(base_iter, buffer_size))
        if not buffered_items: break
        for item in buffered_items:
            yield item


def capture_screenshots(viewer: neuroglancer.Viewer,
                        request_iter: Iterator[CaptureScreenshotRequest],
                        refresh_browser_callback: Callable[[], None],
                        refresh_browser_timeout: int,
                        num_to_prefetch: int = 1) -> None:
    prefetch_buffer = list(itertools.islice(request_iter, num_to_prefetch + 1))
    while prefetch_buffer:
        with viewer.config_state.txn() as s:
            s.show_ui_controls = False
            s.show_panel_borders = False
            del s.prefetch[:]
            for i, request in enumerate(prefetch_buffer[1:]):
                s.prefetch.append(
                    neuroglancer.PrefetchState(state=request.state, priority=num_to_prefetch - i))
            request = prefetch_buffer[0]
            request.config_callback(s)
        viewer.set_state(request.state)
        print('%s [%s] Requesting screenshot' % (
            datetime.datetime.now().strftime('%Y-%m-%dT%H:%M%S.%f'),
            request.description,
        ))
        last_statistics_time = time.time()

        def statistics_callback(statistics):
            nonlocal last_statistics_time
            last_statistics_time = time.time()
            total = statistics.total
            print(
                '%s [%s] Screenshot in progress: %6d/%6d chunks loaded (%10d bytes), %3d downloading'
                % (
                    datetime.datetime.now().strftime('%Y-%m-%dT%H:%M%S.%f'),
                    request.description,
                    total.visible_chunks_gpu_memory,
                    total.visible_chunks_total,
                    total.visible_gpu_memory,
                    total.visible_chunks_downloading,
                ))

        event = threading.Event()
        screenshot = None

        def result_callback(s):
            nonlocal screenshot
            screenshot = s.screenshot
            event.set()

        viewer.async_screenshot(
            result_callback,
            include_depth=request.include_depth,
            statistics_callback=statistics_callback,
        )

        def get_timeout():
            return max(0, last_statistics_time + refresh_browser_timeout - time.time())

        while True:
            if event.wait(get_timeout()):
                break
            if get_timeout() > 0:
                continue
            last_statistics_time = time.time()
            refresh_browser_callback()
        request.response_callback(screenshot)
        del prefetch_buffer[0]
        next_request = next(request_iter, None)
        if next_request is not None:
            prefetch_buffer.append(next_request)


def capture_screenshots_in_parallel(viewers: List[Tuple[neuroglancer.Viewer, Callable[[], None]]],
                                    request_iter: Iterator[CaptureScreenshotRequest],
                                    refresh_browser_timeout: numbers.Number, num_to_prefetch: int,
                                    total_requests: Optional[int] = None,
                                    buffer_size: Optional[int] = None):
    if buffer_size is None:
        if total_requests is None:
            copy_of_requests = list(request_iter)
            total_requests = len(copy_of_requests)
            request_iter = iter(copy_of_requests)
        buffer_size = max(1, total_requests // (len(viewers) * 4))
    request_iter = iter(request_iter)
    threads = []
    buffer_lock = threading.Lock()
    for viewer, refresh_browser_callback in viewers:

        def capture_func(viewer, refresh_browser_callback):
            viewer_request_iter = buffered_iterator(base_iter=request_iter,
                                                    lock=buffer_lock,
                                                    buffer_size=buffer_size)
            capture_screenshots(
                viewer=viewer,
                request_iter=viewer_request_iter,
                num_to_prefetch=num_to_prefetch,
                refresh_browser_timeout=refresh_browser_timeout,
                refresh_browser_callback=refresh_browser_callback,
            )

        t = threading.Thread(target=capture_func, args=(viewer, refresh_browser_callback))
        t.start()
        threads.append(t)
    for t in threads:
        t.join()


class MultiCapturer:
    def __init__(self,
                 shape,
                 include_depth,
                 output,
                 config_callback,
                 num_to_prefetch,
                 checkpoint_interval=60):
        self.include_depth = include_depth
        self.checkpoint_interval = checkpoint_interval
        self.config_callback = config_callback
        self.num_to_prefetch = num_to_prefetch
        self.output = output
        self._processed = set()
        self.state_file = output + '.npz'
        self.temp_state_file = self.state_file + '.tmp'
        self.image_array = np.zeros((shape[1], shape[0], 4), dtype=np.uint8)
        if self.include_depth:
            self.depth_array = np.zeros((shape[1], shape[0]), dtype=np.float32)
        self._load_state()
        self._add_image_lock = threading.Lock()
        self._last_save_time = time.time()
        self._save_state_in_progress = threading.Event()
        self._save_state_in_progress.set()
        self._num_states_processed = 0
        self._start_time = time.time()

    def _load_state(self):
        if not os.path.exists(self.state_file):
            return
        with np.load(self.state_file, allow_pickle=True) as f:
            if self.include_depth:
                self.depth_array = f['depth']
            self.image_array = f['image']
            self._processed = set(f['processed'].ravel()[0])

    def _save_state(self, save_image=False):
        with self._add_image_lock:
            processed = set(self._processed)
        with open(self.temp_state_file, 'wb') as f:
            save_arrays = {
                'image': self.image_array,
                'processed': processed,
            }
            if self.include_depth:
                save_arrays['depth'] = self.depth_array
            np.savez_compressed(f, **save_arrays)
        os.replace(self.temp_state_file, self.state_file)
        if save_image:
            self._save_image()

    def _save_state_async(self, save_image=False):
        print('Starting checkpointing')

        def func():
            try:
                self._save_state()
                print('Done checkpointing')
            finally:
                self._save_state_in_progress.set()

        threading.Thread(target=func, daemon=True).start()

    def _save_image(self):
        im = PIL.Image.fromarray(self.image_array)
        im.save(self.output)

    def _add_image(self, params, screenshot):
        with self._add_image_lock:
            tile_image = screenshot.image_pixels
            tile_selector = np.s_[params['y_offset']:params['y_offset'] + params['tile_height'],
                                  params['x_offset']:params['x_offset'] + params['tile_width']]
            if self.include_depth:
                tile_depth = screenshot.depth_array
                depth_array_part = self.depth_array[tile_selector]
                mask = np.logical_and(np.logical_or(tile_depth != 0, depth_array_part == 0),
                                      tile_depth >= depth_array_part)
                depth_array_part[mask] = tile_depth[mask]
            else:
                mask = Ellipsis
            self.image_array[tile_selector][mask] = tile_image[mask]
            self._processed.add(self._get_description(params))
            self._num_states_processed += 1
            elapsed = time.time() - self._start_time
            print('%4d tiles rendered in %5d seconds: %.1f seconds/tile' %
                  (self._num_states_processed, elapsed, elapsed / self._num_states_processed))

    def _maybe_save_state(self):
        if not self._save_state_in_progress.is_set(): return
        with self._add_image_lock:
            if self._last_save_time + self.checkpoint_interval < time.time():
                self._last_save_time = time.time()
                self._save_state_in_progress.clear()
                self._save_state_async(save_image=False)

    def _get_description(self, params):
        segment_shard = params.get('segment_shard')
        if segment_shard is not None:
            prefix = 'segment_shard=%d ' % (segment_shard, )
        else:
            prefix = ''
        return '%stile_x=%d tile_y=%d' % (prefix, params['tile_x'], params['tile_y'])

    def _make_capture_request(self, params, state):
        description = self._get_description(params)

        if description in self._processed: return None

        def config_callback(s):
            s.viewer_size = (params['tile_width'], params['tile_height'])
            self.config_callback(s)

        def response_callback(screenshot):
            self._add_image(params, screenshot)
            self._maybe_save_state()

        return CaptureScreenshotRequest(state=state,
                                        description=self._get_description(params),
                                        config_callback=config_callback,
                                        response_callback=response_callback,
                                        include_depth=self.include_depth)

    def _get_capture_screenshot_request_iter(self, state_iter):
        for params, state in state_iter:
            request = self._make_capture_request(params, state)
            if request is not None: yield request

    def capture(self, viewers, state_iter, refresh_browser_timeout: int, save_depth: bool, total_requests: int):
        capture_screenshots_in_parallel(
            viewers=viewers,
            request_iter=self._get_capture_screenshot_request_iter(state_iter),
            refresh_browser_timeout=refresh_browser_timeout,
            num_to_prefetch=self.num_to_prefetch,
            total_requests=total_requests)
        if not self._save_state_in_progress.is_set():
            print('Waiting for previous save state to complete')
            self._save_state_in_progress.wait()
        if save_depth:
            self._save_state()
        else:
            self._save_image()
            if os.path.exists(self.state_file):
                os.remove(self.state_file)


def capture_image(viewers, args, state):
    def config_callback(s):
        s.scale_bar_options.scale_factor = args.scale_bar_scale

    segment_shard_size = args.segment_shard_size
    tile_parameters = dict(
        shape=(args.width, args.height),
        tile_shape=(args.tile_width, args.tile_height),
    )
    if segment_shard_size is not None and _should_shard_segments(state, segment_shard_size):
        gen = ShardedTileGenerator(state=state,
                                   segment_shard_size=segment_shard_size,
                                   reverse_bits=args.sort_segments_by_reversed_bits,
                                   **tile_parameters)
        num_states = gen.num_tiles
        state_iter = gen.get_states()
        include_depth = True
    else:
        gen = TileGenerator(**tile_parameters)
        num_states = gen.num_tiles
        state_iter = gen.get_tile_states(state)
        include_depth = False

    capturer = MultiCapturer(
        shape=tile_parameters['shape'],
        include_depth=include_depth,
        output=args.output,
        config_callback=config_callback,
        num_to_prefetch=args.prefetch,
        checkpoint_interval=args.checkpoint_interval,
    )
    num_output_shards = args.num_output_shards
    tiles_per_output_shard = args.tiles_per_output_shard
    output_shard = args.output_shard
    if (output_shard is None) != (num_output_shards is None and tiles_per_output_shard is None):
        raise ValueError(
            '--output-shard must be specified in combination with --num-output-shards or --tiles-per-output-shard'
        )
    if output_shard is not None:
        if num_output_shards is not None:
            if num_output_shards < 1:
                raise ValueError('Invalid --num-output-shards: %d' % (num_output_shards, ))
            states_per_shard = -(-num_states // num_output_shards)
        else:
            if tiles_per_output_shard < 1:
                raise ValueError('Invalid --tiles-per-output-shard: %d' %
                                 (tiles_per_output_shard, ))
            num_output_shards = -(-num_states // tiles_per_output_shard)
            states_per_shard = tiles_per_output_shard
        if output_shard < 0 or output_shard >= num_output_shards:
            raise ValueError('Invalid --output-shard: %d' % (output_shard, ))
        print('Total states: %d, Number of output shards: %d' % (num_states, num_output_shards))
        state_iter = itertools.islice(state_iter, states_per_shard * output_shard,
                                      states_per_shard * (output_shard + 1))
    else:
        states_per_shard = num_states
    capturer.capture(
        viewers=viewers,
        state_iter=state_iter,
        refresh_browser_timeout=args.refresh_browser_timeout,
        save_depth=output_shard is not None,
        total_requests=states_per_shard,
    )


def define_state_modification_args(ap: argparse.ArgumentParser):
    ap.add_argument('--hide-axis-lines',
                    dest='show_axis_lines',
                    action='store_false',
                    help='Override showAxisLines setting in state.')
    ap.add_argument('--hide-default-annotations',
                    action='store_false',
                    dest='show_default_annotations',
                    help='Override showDefaultAnnotations setting in state.')
    ap.add_argument('--projection-scale-multiplier',
                    type=float,
                    help='Multiply projection view scale by specified factor.')
    ap.add_argument('--system-memory-limit',
                    type=int,
                    default=3 * 1024 * 1024 * 1024,
                    help='System memory limit')
    ap.add_argument('--gpu-memory-limit',
                    type=int,
                    default=3 * 1024 * 1024 * 1024,
                    help='GPU memory limit')
    ap.add_argument('--concurrent-downloads', type=int, default=32, help='Concurrent downloads')
    ap.add_argument('--layout', type=str, help='Override layout setting in state.')
    ap.add_argument('--cross-section-background-color',
                    type=str,
                    help='Background color for cross sections.')
    ap.add_argument('--scale-bar-scale', type=float, help='Scale factor for scale bar', default=1)


def apply_state_modifications(state: neuroglancer.ViewerState, args: argparse.Namespace):
    state.selected_layer.visible = False
    state.statistics.visible = False
    if args.layout is not None:
        state.layout = args.layout
    if args.show_axis_lines is not None:
        state.show_axis_lines = args.show_axis_lines
    if args.show_default_annotations is not None:
        state.show_default_annotations = args.show_default_annotations
    if args.projection_scale_multiplier is not None:
        state.projection_scale *= args.projection_scale_multiplier
    if args.cross_section_background_color is not None:
        state.cross_section_background_color = args.cross_section_background_color

    state.gpu_memory_limit = args.gpu_memory_limit
    state.system_memory_limit = args.system_memory_limit
    state.concurrent_downloads = args.concurrent_downloads


def define_viewer_args(ap: argparse.ArgumentParser):
    ap.add_argument('--browser', choices=['chrome', 'firefox'], default='chrome')
    ap.add_argument('--no-webdriver',
                    action='store_true',
                    help='Do not open browser automatically via webdriver.')
    ap.add_argument('--no-headless',
                    dest='headless',
                    action='store_false',
                    help='Use non-headless webdriver.')
    ap.add_argument('--docker-chromedriver',
                    action='store_true',
                    help='Run Chromedriver with options suitable for running inside docker')
    ap.add_argument('--debug-chromedriver',
                    action='store_true',
                    help='Enable debug logging in Chromedriver')
    ap.add_argument('--jobs',
                    '-j',
                    type=int,
                    default=1,
                    help='Number of browsers to use concurrently.  '
                    'This may improve performance at the cost of greater memory usage.  '
                    'On a 64GiB 16 hyperthread machine, --jobs=6 works well.')


def define_size_args(ap: argparse.ArgumentParser):
    ap.add_argument('--width', type=int, default=3840, help='Width in pixels of image.')
    ap.add_argument('--height', type=int, default=2160, help='Height in pixels of image.')


def define_tile_args(ap: argparse.ArgumentParser):
    ap.add_argument(
        '--tile-width',
        type=int,
        default=4096,
        help=
        'Width in pixels of single tile.  If total width is larger, the screenshot will be captured as multiple tiles.'
    )
    ap.add_argument(
        '--tile-height',
        type=int,
        default=4096,
        help=
        'Height in pixels of single tile.  If total height is larger, the screenshot will be captured as multiple tiles.'
    )
    ap.add_argument('--segment-shard-size',
                    type=int,
                    help='Maximum number of segments to render simultaneously.  '
                    'If the number of selected segments exceeds this number, '
                    'multiple passes will be used (transparency not supported).')
    ap.add_argument(
        '--sort-segments-by-reversed-bits',
        action='store_true',
        help=
        'When --segment-shard-size is also specified, normally segment ids are ordered numerically before being partitioned into shards.  If segment ids are spatially correlated, then this can lead to slower and more memory-intensive rendering.  If --sort-segments-by-reversed-bits is specified, segment ids are instead ordered by their bit reversed values, which may avoid the spatial correlation.'
    )


def define_capture_args(ap: argparse.ArgumentParser):
    ap.add_argument('--prefetch', type=int, default=1, help='Number of states to prefetch.')
    ap.add_argument(
        '--refresh-browser-timeout',
        type=int,
        default=60,
        help=
        'Number of seconds without receiving statistics while capturing a screenshot before browser is considered unresponsive.'
    )


@contextlib.contextmanager
def get_viewers(args: argparse.Namespace):
    if args.no_webdriver:
        viewers = [neuroglancer.Viewer() for _ in range(args.jobs)]
        print('Open the following URLs to begin rendering')
        for viewer in viewers:
            print(viewer)

        def refresh_browser_callback():
            print('Browser unresponsive, consider reloading')

        yield [(viewer, refresh_browser_callback) for viewer in viewers]
    else:

        def _make_webdriver():
            webdriver = neuroglancer.webdriver.Webdriver(
                headless=args.headless,
                docker=args.docker_chromedriver,
                debug=args.debug_chromedriver,
                browser=args.browser,
            )

            def refresh_browser_callback():
                print('Browser unresponsive, reloading')
                webdriver.reload_browser()

            return webdriver, refresh_browser_callback

        webdrivers = [_make_webdriver() for _ in range(args.jobs)]
        try:
            yield [(webdriver.viewer, refresh_browser_callback)
                   for webdriver, refresh_browser_callback in webdrivers]
        finally:
            for webdriver, _ in webdrivers:
                try:
                    webdriver.__exit__()
                except:
                    pass


def run(args: argparse.Namespace):
    neuroglancer.cli.handle_server_arguments(args)
    state = args.state
    apply_state_modifications(state, args)
    with get_viewers(args) as viewers:
        capture_image(viewers, args, state)


def main(args=None):
    ap = argparse.ArgumentParser()
    neuroglancer.cli.add_server_arguments(ap)
    neuroglancer.cli.add_state_arguments(ap, required=True)

    ap.add_argument('output', help='Output path of screenshot file in PNG format.')

    ap.add_argument('--output-shard', type=int, help='Output shard to write.')
    output_shard_group = ap.add_mutually_exclusive_group(required=False)
    output_shard_group.add_argument('--num-output-shards',
                                    type=int,
                                    help='Number of output shards.')
    output_shard_group.add_argument('--tiles-per-output-shard',
                                    type=int,
                                    help='Number of tiles per output shard.')
    ap.add_argument('--checkpoint-interval',
                    type=float,
                    default=60,
                    help='Interval in seconds at which to save checkpoints.')

    define_state_modification_args(ap)
    define_viewer_args(ap)
    define_size_args(ap)
    define_tile_args(ap)
    define_capture_args(ap)

    run(ap.parse_args(args))


if __name__ == '__main__':
    main()



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from .server import set_static_content_source, set_server_bind_address, is_server_running, stop
from .viewer import Viewer, UnsynchronizedViewer
from .local_volume import LocalVolume
from .viewer_state import *
from .viewer_config_state import LayerSelectedValues, LayerSelectionState, SegmentIdMapEntry, PrefetchState, ScaleBarOptions
from .equivalence_map import EquivalenceMap
from .url_state import to_url, parse_url, to_json_dump
from .screenshot import ScreenshotSaver
from . import skeleton
from . import server
from . import segment_colors



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, print_function

import concurrent.futures
import json
import multiprocessing
import re
import socket
import sys
import threading
import weakref

import numpy as np
import tornado.httpserver
import tornado.ioloop
import tornado.netutil
import tornado.web

import sockjs.tornado

try:
    # Newer versions of tornado do not have the asynchronous decorator
    from sockjs.tornado.util import asynchronous
except ImportError:
    from tornado.web import asynchronous

from . import local_volume, static
from . import skeleton
from .json_utils import json_encoder_default
from .random_token import make_random_token
from .sockjs_handler import SOCKET_PATH_REGEX, SOCKET_PATH_REGEX_WITHOUT_GROUP, SockJSHandler

INFO_PATH_REGEX = r'^/neuroglancer/info/(?P<token>[^/]+)$'
SKELETON_INFO_PATH_REGEX = r'^/neuroglancer/skeletoninfo/(?P<token>[^/]+)$'

DATA_PATH_REGEX = r'^/neuroglancer/(?P<data_format>[^/]+)/(?P<token>[^/]+)/(?P<scale_key>[^/]+)/(?P<start>[0-9]+(?:,[0-9]+)*)/(?P<end>[0-9]+(?:,[0-9]+)*)$'

SKELETON_PATH_REGEX = r'^/neuroglancer/skeleton/(?P<key>[^/]+)/(?P<object_id>[0-9]+)$'

MESH_PATH_REGEX = r'^/neuroglancer/mesh/(?P<key>[^/]+)/(?P<object_id>[0-9]+)$'

STATIC_PATH_REGEX = r'^/v/(?P<viewer_token>[^/]+)/(?P<path>(?:[a-zA-Z0-9_\-][a-zA-Z0-9_\-.]*)?)$'

ACTION_PATH_REGEX = r'^/action/(?P<viewer_token>[^/]+)$'

global_static_content_source = None

global_server_args = dict(bind_address='127.0.0.1', bind_port=0)

debug = False

class Server(object):
    def __init__(self, ioloop, bind_address='127.0.0.1', bind_port=0):
        self.viewers = weakref.WeakValueDictionary()
        self.token = make_random_token()
        self.executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=multiprocessing.cpu_count())

        self.ioloop = ioloop
        sockjs_router = sockjs.tornado.SockJSRouter(
            SockJSHandler, SOCKET_PATH_REGEX_WITHOUT_GROUP, io_loop=ioloop)
        sockjs_router.neuroglancer_server = self
        def log_function(handler):
            if debug:
                print("%d %s %.2fs" % (handler.get_status(),
                                       handler.request.uri, handler.request.request_time()))

        app = self.app = tornado.web.Application(
            [
                (STATIC_PATH_REGEX, StaticPathHandler, dict(server=self)),
                (INFO_PATH_REGEX, VolumeInfoHandler, dict(server=self)),
                (SKELETON_INFO_PATH_REGEX, SkeletonInfoHandler, dict(server=self)),
                (DATA_PATH_REGEX, SubvolumeHandler, dict(server=self)),
                (SKELETON_PATH_REGEX, SkeletonHandler, dict(server=self)),
                (MESH_PATH_REGEX, MeshHandler, dict(server=self)),
                (ACTION_PATH_REGEX, ActionHandler, dict(server=self)),
            ] + sockjs_router.urls,
            log_function=log_function,
            # Set a large maximum message size to accommodate large screenshot
            # messages.
            websocket_max_message_size=100 * 1024 * 1024)
        http_server = tornado.httpserver.HTTPServer(
            app,
            # Allow very large requests to accommodate large screenshots.
            max_buffer_size=1024**3,
        )
        sockets = tornado.netutil.bind_sockets(port=bind_port, address=bind_address)
        http_server.add_sockets(sockets)
        actual_port = sockets[0].getsockname()[1]

        global global_static_content_source
        if global_static_content_source is None:
            global_static_content_source = static.get_default_static_content_source()

        if bind_address == '0.0.0.0' or bind_address == '::':
            hostname = socket.getfqdn()
        else:
            hostname = bind_address

        self.server_url = 'http://%s:%s' % (hostname, actual_port)

    def get_volume(self, key):
        dot_index = key.find('.')
        if dot_index == -1:
            return None
        viewer_token = key[:dot_index]
        volume_token = key[dot_index+1:]
        viewer = self.viewers.get(viewer_token)
        if viewer is None:
            return None
        return viewer.volume_manager.volumes.get(volume_token)


class BaseRequestHandler(tornado.web.RequestHandler):
    def initialize(self, server):
        self.server = server

class StaticPathHandler(BaseRequestHandler):
    def get(self, viewer_token, path):
        if viewer_token != self.server.token and viewer_token not in self.server.viewers:
            self.send_error(404)
            return
        try:
            data, content_type = global_static_content_source.get(path)
        except ValueError as e:
            self.send_error(404, message=e.args[0])
            return
        self.set_header('Content-type', content_type)
        self.finish(data)

class ActionHandler(BaseRequestHandler):
    def post(self, viewer_token):
        viewer = self.server.viewers.get(viewer_token)
        if viewer is None:
            self.send_error(404)
            return
        action = json.loads(self.request.body)
        self.server.ioloop.add_callback(viewer.actions.invoke, action['action'], action['state'])
        self.finish('')

class VolumeInfoHandler(BaseRequestHandler):
    def get(self, token):
        vol = self.server.get_volume(token)
        if vol is None or not isinstance(vol, local_volume.LocalVolume):
            self.send_error(404)
            return
        self.finish(json.dumps(vol.info(), default=json_encoder_default).encode())

class SkeletonInfoHandler(BaseRequestHandler):
    def get(self, token):
        vol = self.server.get_volume(token)
        if vol is None or not isinstance(vol, skeleton.SkeletonSource):
            self.send_error(404)
            return
        self.finish(json.dumps(vol.info(), default=json_encoder_default).encode())

class SubvolumeHandler(BaseRequestHandler):
    @asynchronous
    def get(self, data_format, token, scale_key, start, end):
        start_pos = np.array(start.split(','), dtype=np.int64)
        end_pos = np.array(end.split(','), dtype=np.int64)
        vol = self.server.get_volume(token)
        if vol is None or not isinstance(vol, local_volume.LocalVolume):
            self.send_error(404)
            return

        def handle_subvolume_result(f):
            try:
                data, content_type = f.result()
            except ValueError as e:
                self.send_error(400, message=e.args[0])
                return

            self.set_header('Content-type', content_type)
            self.finish(data)

        self.server.executor.submit(
            vol.get_encoded_subvolume,
            data_format=data_format, start=start_pos, end=end_pos, scale_key=scale_key).add_done_callback(
                lambda f: self.server.ioloop.add_callback(lambda: handle_subvolume_result(f)))


class MeshHandler(BaseRequestHandler):
    @asynchronous
    def get(self, key, object_id):
        object_id = int(object_id)
        vol = self.server.get_volume(key)
        if vol is None or not isinstance(vol, local_volume.LocalVolume):
            self.send_error(404)
            return

        def handle_mesh_result(f):
            try:
                encoded_mesh = f.result()
            except local_volume.MeshImplementationNotAvailable:
                self.send_error(501, message='Mesh implementation not available')
                return
            except local_volume.MeshesNotSupportedForVolume:
                self.send_error(405, message='Meshes not supported for volume')
                return
            except local_volume.InvalidObjectIdForMesh:
                self.send_error(404, message='Mesh not available for specified object id')
                return
            except ValueError as e:
                self.send_error(400, message=e.args[0])
                return

            self.set_header('Content-type', 'application/octet-stream')
            self.finish(encoded_mesh)

        self.server.executor.submit(vol.get_object_mesh, object_id).add_done_callback(
            lambda f: self.server.ioloop.add_callback(lambda: handle_mesh_result(f)))


class SkeletonHandler(BaseRequestHandler):
    @asynchronous
    def get(self, key, object_id):
        object_id = int(object_id)
        vol = self.server.get_volume(key)
        if vol is None or not isinstance(vol, skeleton.SkeletonSource):
            self.send_error(404)

        def handle_result(f):
            try:
                encoded_skeleton = f.result()
            except:
                self.send_error(500, message=e.args[0])
                return
            if encoded_skeleton is None:
                self.send_error(404, message='Skeleton not available for specified object id')
                return
            self.set_header('Content-type', 'application/octet-stream')
            self.finish(encoded_skeleton)

        def get_encoded_skeleton(skeletons, object_id):
            skeleton = skeletons.get_skeleton(object_id)
            if skeleton is None:
                return None
            return skeleton.encode(skeletons)

        self.server.executor.submit(
            get_encoded_skeleton, vol, object_id).add_done_callback(
                lambda f: self.server.ioloop.add_callback(lambda: handle_result(f)))


global_server = None


def set_static_content_source(*args, **kwargs):
    global global_static_content_source
    global_static_content_source = static.get_static_content_source(*args, **kwargs)


def set_server_bind_address(bind_address='127.0.0.1', bind_port=0):
    global global_server_args
    global_server_args = dict(bind_address=bind_address, bind_port=bind_port)


def is_server_running():
    return global_server is not None


def stop():
    """Stop the server, invalidating any viewer URLs.

    This allows any previously-referenced data arrays to be garbage collected if there are no other
    references to them.
    """
    global global_server
    if global_server is not None:
        ioloop = global_server.ioloop
        def stop_ioloop():
            ioloop.stop()
        global_server.ioloop.add_callback(stop_ioloop)
        global_server = None


def get_server_url():
    return global_server.server_url


_global_server_lock = threading.Lock()


def start():
    global global_server
    with _global_server_lock:
        if global_server is not None: return

        # Workaround https://bugs.python.org/issue37373
        # https://www.tornadoweb.org/en/stable/index.html#installation
        if sys.platform == 'win32' and sys.version_info >= (3, 8):
            import asyncio
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

        done = threading.Event()

        def start_server():
            global global_server
            ioloop = tornado.ioloop.IOLoop()
            ioloop.make_current()
            global_server = Server(ioloop=ioloop, **global_server_args)
            done.set()
            ioloop.start()
            ioloop.close()

        thread = threading.Thread(target=start_server)
        thread.daemon = True
        thread.start()
        done.wait()


def register_viewer(viewer):
    start()
    global_server.viewers[viewer.token] = viewer

def defer_callback(callback, *args, **kwargs):
    """Register `callback` to run in the server event loop thread."""
    start()
    global_server.ioloop.add_callback(lambda: callback(*args, **kwargs))



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import zlib

import numpy as np
from PIL import Image


def encode_jpeg(subvol):
    shape = subvol.shape
    reshaped = subvol.reshape(shape[0] * shape[1], shape[2])
    img = Image.fromarray(reshaped)
    f = io.BytesIO()
    img.save(f, "JPEG")
    return f.getvalue()


def encode_npz(subvol):
    fileobj = io.BytesIO()
    np.save(fileobj, np.asfortranarray(subvol))
    cdz = zlib.compress(fileobj.getvalue())
    return cdz


def encode_raw(subvol):
    return subvol.tostring('C')



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import collections
import json
import re

from six.moves import urllib

from . import viewer_state
from .json_utils import json_encoder_default
from .json_wrappers import to_json

SINGLE_QUOTE_STRING_PATTERN = u'(\'(?:[^\'\\\\]|(?:\\\\.))*\')'
DOUBLE_QUOTE_STRING_PATTERN = u'("(?:[^"\\\\]|(?:\\\\.))*")'
SINGLE_OR_DOUBLE_QUOTE_STRING_PATTERN = SINGLE_QUOTE_STRING_PATTERN + u'|' + DOUBLE_QUOTE_STRING_PATTERN
DOUBLE_OR_SINGLE_QUOTE_STRING_PATTERN = DOUBLE_QUOTE_STRING_PATTERN + u'|' + SINGLE_QUOTE_STRING_PATTERN


DOUBLE_QUOTE_PATTERN = u'^((?:[^"\'\\\\]|(?:\\\\.))*)"'
SINGLE_QUOTE_PATTERN = u'^((?:[^"\'\\\\]|(?:\\\\.))*)\''


def _convert_string_literal(x, quote_initial, quote_replace, quote_search):
    if len(x) >= 2 and x[0] == quote_initial and x[-1] == quote_initial:
        inner = x[1:-1]
        s = quote_replace
        while inner:
            m = re.search(quote_search, inner)
            if m is None:
                s += inner
                break
            s += m.group(1)
            s += u'\\'
            s += quote_replace
            inner = inner[m.end():]
        s += quote_replace
        return s
    return x


def _convert_json_helper(x, desired_comma_char, desired_quote_char):
    comma_search = u'[&_,]'
    if desired_quote_char == u'"':
        quote_initial = u'\''
        quote_search = DOUBLE_QUOTE_PATTERN
        string_literal_pattern = SINGLE_OR_DOUBLE_QUOTE_STRING_PATTERN
    else:
        quote_initial = u'"'
        quote_search = SINGLE_QUOTE_PATTERN
        string_literal_pattern = DOUBLE_OR_SINGLE_QUOTE_STRING_PATTERN
    s = u''
    while x:
        m = re.search(string_literal_pattern, x)
        if m is None:
            before = x
            x = u''
            replacement = u''
        else:
            before = x[:m.start()]
            x = x[m.end():]
            original_string = m.group(1)
            if original_string is not None:
                replacement = _convert_string_literal(original_string, quote_initial, desired_quote_char, quote_search)
            else:
                replacement = m.group(2)
        s += re.sub(comma_search, desired_comma_char, before)
        s += replacement
    return s


def url_safe_to_json(x):
    return _convert_json_helper(x, u',', u'"')

def json_to_url_safe(x):
    return _convert_json_helper(x, u'_', u'\'')

def url_fragment_to_json(fragment_value):
    unquoted = urllib.parse.unquote(fragment_value)
    if unquoted.startswith('!'):
        unquoted = unquoted[1:]
    return url_safe_to_json(unquoted)


def parse_url_fragment(fragment_value):
    json_string = url_fragment_to_json(fragment_value)
    return viewer_state.ViewerState(
        json.loads(json_string, object_pairs_hook=collections.OrderedDict))


def parse_url(url):
    result = urllib.parse.urlparse(url)
    return parse_url_fragment(result.fragment)

def to_url_fragment(state):
    json_string = json.dumps(to_json(state), separators=(u',', u':'), default=json_encoder_default)
    return urllib.parse.quote(json_string, safe=u'~@#$&()*!+=:;,.?/\'')

default_neuroglancer_url = u'https://neuroglancer-demo.appspot.com'

def to_url(state, prefix=default_neuroglancer_url):
    return u'%s#!%s' % (prefix, to_url_fragment(state))

def to_json_dump(state, indent=None, separators=None):
    return json.dumps(to_json(state), separators=separators,
                      indent=indent,
                      default=json_encoder_default)
